@@ -32,6 +32,8 @@
 #include <stdio.h>
 #include <string.h>
 
+static void handle_get_items (DBusPendingCall *pending, void *user_data);
+
 static DBusConnection *bus = NULL;
 static GHashTable *apps = NULL;
 static GHashTable *live_refs = NULL;
@@ -133,11 +135,51 @@ static gboolean atspi_inited = FALSE;
 
 static GHashTable *app_hash = NULL;
 
+static void
+handle_get_bus_address (DBusPendingCall *pending, void *user_data)
+{
+  AtspiApplication *app = user_data;
+  DBusMessage *reply = dbus_pending_call_steal_reply (pending);
+  DBusMessage *message;
+  const char *address;
+  DBusPendingCall *new_pending;
+
+  if (dbus_message_get_type (reply) == DBUS_MESSAGE_TYPE_METHOD_RETURN)
+  {
+    if (dbus_message_get_args (reply, NULL, DBUS_TYPE_STRING, &address,
+                               DBUS_TYPE_INVALID))
+    {
+      DBusError error;
+      dbus_error_init (&error);
+      DBusConnection *bus = dbus_connection_open (address, &error);
+      if (bus)
+      {
+        if (app->bus)
+          dbus_connection_unref (app->bus);
+        app->bus = bus;
+      }
+    }
+  }
+  dbus_message_unref (reply);
+  dbus_pending_call_unref (pending);
+
+  message = dbus_message_new_method_call (app->bus_name,
+                                          "/org/a11y/atspi/cache",
+                                          atspi_interface_cache, "GetItems");
+
+   dbus_connection_send_with_reply (app->bus, message, &new_pending, 2000);
+  dbus_pending_call_set_notify (new_pending, handle_get_items, app, NULL);
+  dbus_message_unref (message);
+}
+
 static AtspiApplication *
 get_application (const char *bus_name)
 {
   AtspiApplication *app = NULL;
   char *bus_name_dup;
+  DBusMessage *message;
+  DBusError error;
+  DBusPendingCall *pending = NULL;
 
   if (!app_hash)
   {
@@ -153,19 +195,29 @@ get_application (const char *bus_name)
   if (!app) return NULL;
   app->bus_name = bus_name_dup;
   app->hash = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, g_object_unref);
+  app->bus = dbus_connection_ref (_atspi_bus ());
   g_hash_table_insert (app_hash, bus_name_dup, app);
+  dbus_error_init (&error);
+  message = dbus_message_new_method_call (bus_name, atspi_path_root,
+                                          atspi_interface_application, "GetApplicationBusAddress");
+
+   dbus_connection_send_with_reply (app->bus, message, &pending, 2000);
+  dbus_pending_call_set_notify (pending, handle_get_bus_address, app, NULL);
+  dbus_message_unref (message);
   return app;
 }
 
 static AtspiAccessible *
 ref_accessible (const char *app_name, const char *path)
 {
-  AtspiApplication *app = get_application (app_name);
+  AtspiApplication *app;
   AtspiAccessible *a;
 
   if (!strcmp (path, ATSPI_DBUS_PATH_NULL))
     return NULL;
 
+  app = get_application (app_name);
+
   if (!strcmp (path, "/org/a11y/atspi/accessible/root"))
   {
     if (!app->root)
@@ -184,7 +236,7 @@ ref_accessible (const char *app_name, const char *path)
   a = atspi_accessible_new (app, path);
   if (!a)
     return NULL;
-  g_hash_table_insert (app->hash, a->parent.path, a);
+  g_hash_table_insert (app->hash, g_strdup (a->parent.path), a);
   g_object_ref (a);	/* for the hash */
   return a;
 }
@@ -206,7 +258,7 @@ ref_hyperlink (const char *app_name, const char *path)
   hyperlink = atspi_hyperlink_new (app, path);
   if (!hyperlink)
     return NULL;
-  g_hash_table_insert (app->hash, hyperlink->parent.path, hyperlink);
+  g_hash_table_insert (app->hash, g_strdup (hyperlink->parent.path), hyperlink);
   /* TODO: This should be a weak ref */
   g_object_ref (hyperlink);	/* for the hash */
   return hyperlink;
@@ -237,24 +289,21 @@ handle_remove_accessible (DBusConnection *bus, DBusMessage *message, void *user_
 
   if (strcmp (signature, "(so)") != 0)
   {
-    g_warning ("at-spi: Unknown signature %s for RemoveAccessible", signature);
+    g_warning (_("AT-SPI: Unknown signature %s for RemoveAccessible"), signature);
     return DBUS_HANDLER_RESULT_HANDLED;
   }
 
   dbus_message_iter_init (message, &iter);
   dbus_message_iter_recurse (&iter, &iter_struct);
   dbus_message_iter_get_basic (&iter_struct, &sender);
+  dbus_message_iter_next (&iter_struct);
   dbus_message_iter_get_basic (&iter_struct, &path);
   app = get_application (sender);
   a = ref_accessible (sender, path);
   if (!a)
     return DBUS_HANDLER_RESULT_HANDLED;
-  if (a->accessible_parent && g_list_find (a->accessible_parent->children, a))
-  {
-    a->accessible_parent->children = g_list_remove (a->accessible_parent->children, a);
-    g_object_unref (a);
-  }
-  g_hash_table_remove (app->hash, app->bus_name);
+  g_object_run_dispose (G_OBJECT (a));
+  g_hash_table_remove (app->hash, a->parent.path);
   g_object_unref (a);	/* unref our own ref */
   return DBUS_HANDLER_RESULT_HANDLED;
 }
@@ -278,7 +327,7 @@ add_app_to_desktop (AtspiAccessible *a, const char *bus_name)
   }
   else
   {
-    g_warning ("Error calling getRoot for %s: %s", bus_name, error.message);
+    g_warning (_("AT-SPI: Error calling getRoot for %s: %s"), bus_name, error.message);
   }
   return FALSE;
 }
@@ -321,7 +370,6 @@ remove_app_from_desktop (AtspiAccessible *a, const char *bus_name)
   }
   if (!l)
   {
-    g_warning ("Removing unregistered app %s; doing nothing\n", bus_name);
     return FALSE;
   }
   send_children_changed (a, child, FALSE);
@@ -355,8 +403,6 @@ add_accessible_from_iter (DBusMessageIter *iter)
   AtspiAccessible *accessible;
   const char *name, *description;
   dbus_uint32_t role;
-  dbus_uint32_t *states;
-  int count;
 
   dbus_message_iter_recurse (iter, &iter_struct);
 
@@ -371,9 +417,16 @@ add_accessible_from_iter (DBusMessageIter *iter)
 
   /* get parent */
   get_reference_from_iter (&iter_struct, &app_name, &path);
+  if (accessible->accessible_parent)
+    g_object_unref (accessible->accessible_parent);
   accessible->accessible_parent = ref_accessible (app_name, path);
 
   /* Get children */
+  while (accessible->children)
+  {
+    g_object_unref (accessible->children->data);
+    accessible->children = g_list_remove (accessible->children, accessible->children->data);
+  }
   dbus_message_iter_recurse (&iter_struct, &iter_array);
   while (dbus_message_iter_get_arg_type (&iter_array) != DBUS_TYPE_INVALID)
   {
@@ -385,26 +438,13 @@ add_accessible_from_iter (DBusMessageIter *iter)
   }
 
   /* interfaces */
-  accessible->interfaces = 0;
   dbus_message_iter_next (&iter_struct);
-  dbus_message_iter_recurse (&iter_struct, &iter_array);
-  while (dbus_message_iter_get_arg_type (&iter_array) != DBUS_TYPE_INVALID)
-  {
-    const char *iface;
-    gint n;
-    dbus_message_iter_get_basic (&iter_array, &iface);
-    if (!strcmp (iface, "org.freedesktop.DBus.Introspectable")) continue;
-    n = _atspi_get_iface_num (iface);
-    if (n == -1)
-    {
-      g_warning ("Unknown interface %s", iface);
-    }
-    else accessible->interfaces |= (1 << n);
-    dbus_message_iter_next (&iter_array);
-  }
+  _atspi_dbus_set_interfaces (accessible, &iter_struct);
   dbus_message_iter_next (&iter_struct);
 
   /* name */
+  if (accessible->name)
+    g_free (accessible->name);
   dbus_message_iter_get_basic (&iter_struct, &name);
   accessible->name = g_strdup (name);
   dbus_message_iter_next (&iter_struct);
@@ -415,49 +455,44 @@ add_accessible_from_iter (DBusMessageIter *iter)
   dbus_message_iter_next (&iter_struct);
 
   /* description */
+  if (accessible->description)
+    g_free (accessible->description);
   dbus_message_iter_get_basic (&iter_struct, &description);
   accessible->description = g_strdup (description);
   dbus_message_iter_next (&iter_struct);
 
-  dbus_message_iter_recurse (&iter_struct, &iter_array);
-  dbus_message_iter_get_fixed_array (&iter_array, &states, &count);
-  if (count != 2)
-  {
-    g_warning ("at-spi: expected 2 values in states array; got %d\n", count);
-    accessible->states = _atspi_state_set_new_internal (accessible, 0);
-  }
-  else
-  {
-    guint64 val = ((guint64)states [1]) << 32;
-    val += states [0];
-    accessible->states = _atspi_state_set_new_internal (accessible, val);
-  }
+  _atspi_dbus_set_state (accessible, &iter_struct);
   dbus_message_iter_next (&iter_struct);
 
+  accessible->cached_properties |= ATSPI_CACHE_NAME | ATSPI_CACHE_ROLE | ATSPI_CACHE_PARENT;
+  if (!atspi_state_set_contains (accessible->states,
+                                       ATSPI_STATE_MANAGES_DESCENDANTS))
+    accessible->cached_properties |= ATSPI_CACHE_CHILDREN;
+
   /* This is a bit of a hack since the cache holds a ref, so we don't need
    * the one provided for us anymore */
   g_object_unref (accessible);
 }
 
 static void
-add_accessibles (const char *app_name)
+handle_get_items (DBusPendingCall *pending, void *user_data)
 {
-  DBusError error;
-  DBusMessage *message, *reply;
+  AtspiApplication *app = user_data;
+  DBusMessage *reply = dbus_pending_call_steal_reply (pending);
   DBusMessageIter iter, iter_array;
 
-  AtspiApplication *app = get_application (app_name);
-  /* TODO: Move this functionality into app initializer? */
-  dbus_error_init (&error);
-  message = dbus_message_new_method_call (app_name, "/org/a11y/atspi/cache", atspi_interface_cache, "GetItems");
-  reply = _atspi_dbus_send_with_reply_and_block (message);
-  if (!reply || strcmp (dbus_message_get_signature (reply), "a((so)(so)(so)a(so)assusau)") != 0)
+  if (dbus_message_get_type (reply) == DBUS_MESSAGE_TYPE_ERROR)
   {
-    g_warning ("at-spi: Error in GetItems");
+    const char *sender = dbus_message_get_sender (reply);
+    const char *error = NULL;
+    dbus_message_get_args (reply, NULL, DBUS_TYPE_STRING, &error,
+                           DBUS_TYPE_INVALID);
+    g_warning (_("AT-SPI: Error in GetItems, sender=%s, error=%s"), sender, error);
+    dbus_message_unref (reply);
+    dbus_pending_call_unref (pending);
     return;
-    if (reply)
-      dbus_message_unref (reply);
   }
+
   dbus_message_iter_init (reply, &iter);
   dbus_message_iter_recurse (&iter, &iter_array);
   while (dbus_message_iter_get_arg_type (&iter_array) != DBUS_TYPE_INVALID)
@@ -466,6 +501,7 @@ add_accessibles (const char *app_name)
     dbus_message_iter_next (&iter_array);
   }
   dbus_message_unref (reply);
+  dbus_pending_call_unref (pending);
 }
 
 /* TODO: Do we stil need this function? */
@@ -478,6 +514,7 @@ ref_accessible_desktop (AtspiApplication *app)
   gint i;
   DBusMessage *message, *reply;
   DBusMessageIter iter, iter_array;
+  gchar *bus_name_dup;
 
   if (desktop)
   {
@@ -513,10 +550,15 @@ ref_accessible_desktop (AtspiApplication *app)
   {
     const char *app_name, *path;
     get_reference_from_iter (&iter_array, &app_name, &path);
-    add_accessibles (app_name);
     add_app_to_desktop (desktop, app_name);
   }
   dbus_message_unref (reply);
+
+  /* Record the alternate name as an alias for org.a11y.atspi.Registry */
+  bus_name_dup = g_strdup (dbus_message_get_sender (reply));
+  if (bus_name_dup)
+    g_hash_table_insert (app_hash, bus_name_dup, app);
+
   return desktop;
 }
 
@@ -537,8 +579,12 @@ _atspi_dbus_return_accessible_from_message (DBusMessage *message)
 {
   DBusMessageIter iter;
   AtspiAccessible *retval = NULL;
-  const char *signature = dbus_message_get_signature (message);
-   
+  const char *signature;
+
+  if (!message)
+    return NULL;
+
+  signature = dbus_message_get_signature (message);
   if (!strcmp (signature, "(so)"))
   {
     dbus_message_iter_init (message, &iter);
@@ -546,7 +592,7 @@ _atspi_dbus_return_accessible_from_message (DBusMessage *message)
   }
   else
   {
-    g_warning ("Atspi: Called _atspi_dbus_return_accessible_from_message with strange signature %s", signature);
+    g_warning (_("AT-SPI: Called _atspi_dbus_return_accessible_from_message with strange signature %s"), signature);
   }
   dbus_message_unref (message);
   return retval;
@@ -575,7 +621,7 @@ _atspi_dbus_return_hyperlink_from_message (DBusMessage *message)
   }
   else
   {
-    g_warning ("Atspi: Called _atspi_dbus_return_hyperlink_from_message with strange signature %s", signature);
+    g_warning (_("AT-SPI: Called _atspi_dbus_return_hyperlink_from_message with strange signature %s"), signature);
   }
   dbus_message_unref (message);
   return retval;
@@ -602,7 +648,8 @@ handle_add_accessible (DBusConnection *bus, DBusMessage *message, void *user_dat
 
   if (strcmp (dbus_message_get_signature (message), cache_signal_type) != 0)
   {
-    g_warning ("atspi: AddAccessible with unknown signature %s\n", dbus_message_get_signature (message));
+    g_warning (_("AT-SPI: AddAccessible with unknown signature %s\n"),
+               dbus_message_get_signature (message));
     return;
   }
 
@@ -610,6 +657,91 @@ handle_add_accessible (DBusConnection *bus, DBusMessage *message, void *user_dat
   add_accessible_from_iter (&iter);
 }
 
+typedef struct
+{
+  DBusConnection *bus;
+  DBusMessage *message;
+  void *data;
+} BusDataClosure;
+
+static guint process_deferred_messages_id = -1;
+
+static void
+process_deferred_message (BusDataClosure *closure)
+{
+  int type = dbus_message_get_type (closure->message);
+  const char *interface = dbus_message_get_interface (closure->message);
+  const char *member = dbus_message_get_member (closure->message); 
+  dbus_uint32_t v;
+  char *bus_name;
+
+  if (type == DBUS_MESSAGE_TYPE_SIGNAL &&
+      !strncmp (interface, "org.a11y.atspi.Event.", 21))
+  {
+    atspi_dbus_handle_event (closure->bus, closure->message, closure->data);
+  }
+  if (dbus_message_is_method_call (closure->message, atspi_interface_device_event_listener, "NotifyEvent"))
+  {
+    atspi_dbus_handle_DeviceEvent (closure->bus,
+                                   closure->message, closure->data);
+  }
+  if (dbus_message_is_signal (closure->message, atspi_interface_cache, "AddAccessible"))
+  {
+    handle_add_accessible (closure->bus, closure->message, closure->data);
+  }
+  if (dbus_message_is_signal (closure->message, atspi_interface_cache, "RemoveAccessible"))
+  {
+    handle_remove_accessible (closure->bus, closure->message, closure->data);
+  }
+}
+
+static GList *deferred_messages = NULL;
+
+gboolean
+_atspi_process_deferred_messages (gpointer data)
+{
+  static int in_process_deferred_messages = 0;
+
+  if (in_process_deferred_messages)
+    return;
+  in_process_deferred_messages = 1;
+  while (deferred_messages != NULL)
+  {
+    BusDataClosure *closure = deferred_messages->data;
+    process_deferred_message (closure);
+    deferred_messages = g_list_remove (deferred_messages, closure);
+    dbus_message_unref (closure->message);
+    dbus_connection_unref (closure->bus);
+    g_free (closure);
+  }
+  /* If data is NULL, assume that we were called from GLib */
+  if (!data)
+    process_deferred_messages_id = -1;
+  in_process_deferred_messages = 0;
+  return FALSE;
+}
+
+static DBusHandlerResult
+defer_message (DBusConnection *connection, DBusMessage *message, void *user_data)
+{
+  BusDataClosure *closure = g_new (BusDataClosure, 1);
+  GList *new_list;
+
+  if (!closure)
+    return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+  closure->bus = dbus_connection_ref (bus);
+  closure->message = dbus_message_ref (message);
+  closure->data = user_data;
+
+  new_list = g_list_append (deferred_messages, closure);
+  if (new_list)
+    deferred_messages = new_list;
+
+  if (process_deferred_messages_id == -1)
+    process_deferred_messages_id = g_idle_add (_atspi_process_deferred_messages, NULL);
+  return DBUS_HANDLER_RESULT_HANDLED;
+}
+
 static DBusHandlerResult
 atspi_dbus_filter (DBusConnection *bus, DBusMessage *message, void *data)
 {
@@ -622,21 +754,20 @@ atspi_dbus_filter (DBusConnection *bus, DBusMessage *message, void *data)
   if (type == DBUS_MESSAGE_TYPE_SIGNAL &&
       !strncmp (interface, "org.a11y.atspi.Event.", 21))
   {
-    return atspi_dbus_handle_event (bus, message, data);
+    return defer_message (bus, message, data);
   }
   if (dbus_message_is_method_call (message, atspi_interface_device_event_listener, "NotifyEvent"))
   {
-    return atspi_dbus_handle_DeviceEvent (bus, message, data);
+    return defer_message (bus, message, data);
   }
   if (dbus_message_is_signal (message, atspi_interface_cache, "AddAccessible"))
   {
-    return handle_add_accessible (bus, message, data);
+    return defer_message (bus, message, data);
   }
   if (dbus_message_is_signal (message, atspi_interface_cache, "RemoveAccessible"))
   {
-    return handle_remove_accessible (bus, message, data);
+    return defer_message (bus, message, data);
   }
-  /* TODO: Handle ChildrenChanged, StateChanged, PropertyChanged */
   return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
 }
 
@@ -707,7 +838,7 @@ get_accessibility_bus ()
   bridge_display = XOpenDisplay (spi_display_name ());
   if (!bridge_display)
     {
-      g_warning ("AT_SPI: Could not get the display\n");
+      g_warning (_("AT-SPI: Could not get the display\n"));
       return NULL;
     }
 
@@ -718,17 +849,18 @@ get_accessibility_bus ()
                       (long) BUFSIZ, False,
                       (Atom) 31, &actual_type, &actual_format,
                       &nitems, &leftover, &data);
+  XCloseDisplay (bridge_display);
 
   dbus_error_init (&error);
 
   if (data == NULL)
     {
       g_warning
-        ("AT-SPI: Accessibility bus not found - Using session bus.\n");
+        (_("AT-SPI: Accessibility bus not found - Using session bus.\n"));
       bus = dbus_bus_get (DBUS_BUS_SESSION, &error);
       if (!bus)
         {
-          g_warning ("AT-SPI: Couldn't connect to bus: %s\n", error.message);
+          g_warning (_("AT-SPI: Couldn't connect to bus: %s\n"), error.message);
           return NULL;
         }
     }
@@ -737,14 +869,14 @@ get_accessibility_bus ()
       bus = dbus_connection_open (data, &error);
       if (!bus)
         {
-          g_warning ("AT-SPI: Couldn't connect to bus: %s\n", error.message);
+          g_warning (_("AT-SPI: Couldn't connect to bus: %s\n"), error.message);
           return NULL;
         }
       else
         {
           if (!dbus_bus_register (bus, &error))
             {
-              g_warning ("AT-SPI: Couldn't register with bus: %s\n", error.message);
+              g_warning (_("AT-SPI: Couldn't register with bus: %s\n"), error.message);
               return NULL;
             }
         }
@@ -788,6 +920,7 @@ atspi_init (void)
   dbus_bus_register (bus, &error);
   dbus_connection_setup_with_g_main(bus, g_main_context_default());
   dbus_connection_add_filter (bus, atspi_dbus_filter, NULL, NULL);
+  dbind_set_timeout (1000);
   match = g_strdup_printf ("type='signal',interface='%s',member='AddAccessible'", atspi_interface_cache);
   dbus_error_init (&error);
   dbus_bus_add_match (bus, match, &error);
@@ -879,10 +1012,20 @@ _atspi_dbus_call (gpointer obj, const char *interface, const char *method, GErro
   DBusError err;
   AtspiObject *aobj = ATSPI_OBJECT (obj);
 
+  if (!aobj->app || !aobj->app->bus)
+  {
+    g_set_error_literal (error, ATSPI_ERROR, ATSPI_ERROR_APPLICATION_GONE,
+                          _("The application no longer exists"));
+    return FALSE;
+  }
+
   va_start (args, type);
   dbus_error_init (&err);
-  retval = dbind_method_call_reentrant_va (_atspi_bus(), aobj->app->bus_name, aobj->path, interface, method, &err, type, args);
+  retval = dbind_method_call_reentrant_va (aobj->app->bus, aobj->app->bus_name,
+                                           aobj->path, interface, method, &err,
+                                           type, args);
   va_end (args);
+  _atspi_process_deferred_messages ((gpointer)TRUE);
   if (dbus_error_is_set (&err))
   {
     /* TODO: Set gerror */
@@ -921,6 +1064,13 @@ _atspi_dbus_call_partial_va (gpointer obj,
 
   dbus_error_init (&err);
 
+  if (!aobj->app || !aobj->app->bus)
+  {
+    g_set_error_literal (error, ATSPI_ERROR, ATSPI_ERROR_APPLICATION_GONE,
+                          _("The application no longer exists"));
+    goto out;
+  }
+
     msg = dbus_message_new_method_call (aobj->app->bus_name, aobj->path, interface, method);
     if (!msg)
         goto out;
@@ -929,9 +1079,12 @@ _atspi_dbus_call_partial_va (gpointer obj,
     dbus_message_iter_init_append (msg, &iter);
     dbind_any_marshal_va (&iter, &p, args);
 
-    reply = dbind_send_and_allow_reentry (_atspi_bus(), msg, &err);
+    reply = dbind_send_and_allow_reentry (aobj->app->bus, msg, &err);
 out:
   va_end (args);
+  if (msg)
+    dbus_message_unref (msg);
+  _atspi_process_deferred_messages ((gpointer)TRUE);
   if (dbus_error_is_set (&err))
   {
     /* TODO: Set gerror */
@@ -950,7 +1103,14 @@ _atspi_dbus_get_property (gpointer obj, const char *interface, const char *name,
   AtspiObject *aobj = ATSPI_OBJECT (obj);
 
   if (!aobj)
-    return NULL;
+    return FALSE;
+
+  if (!aobj->app || !aobj->app->bus)
+  {
+    g_set_error_literal (error, ATSPI_ERROR, ATSPI_ERROR_APPLICATION_GONE,
+                          _("The application no longer exists"));
+    return FALSE;
+  }
 
   message = dbus_message_new_method_call (aobj->app->bus_name,
                                           aobj->path,
@@ -959,22 +1119,38 @@ _atspi_dbus_get_property (gpointer obj, const char *interface, const char *name,
   if (!message)
   {
     // TODO: throw exception
-    goto done;
+    return FALSE;
   }
   dbus_message_append_args (message, DBUS_TYPE_STRING, &interface, DBUS_TYPE_STRING, &name, DBUS_TYPE_INVALID);
   dbus_error_init (&err);
-  reply = dbus_connection_send_with_reply_and_block (_atspi_bus(), message, 1000, &err);
+  reply = dbind_send_and_allow_reentry (aobj->app->bus, message, &err);
   dbus_message_unref (message);
+  _atspi_process_deferred_messages ((gpointer)TRUE);
   if (!reply)
   {
     // TODO: throw exception
     goto done;
   }
+
+  if (dbus_message_get_type (reply) == DBUS_MESSAGE_TYPE_ERROR)
+  {
+    const char *err;
+    dbus_message_get_args (message, NULL, DBUS_TYPE_STRING, &err, DBUS_TYPE_INVALID);
+    if (err)
+      g_set_error_literal (error, ATSPI_ERROR, ATSPI_ERROR_IPC, err);
+    goto done;
+  }
+
   dbus_message_iter_init (reply, &iter);
+  if (dbus_message_iter_get_arg_type (&iter) != 'v')
+  {
+    g_warning (_("AT-SPI: expected a variant when fetching %s from interface %s; got %s\n"), name, interface, dbus_message_get_signature (reply));
+    goto done;
+  }
   dbus_message_iter_recurse (&iter, &iter_variant);
   if (dbus_message_iter_get_arg_type (&iter_variant) != type[0])
   {
-    g_warning ("atspi_dbus_get_property: Wrong type: expected %s, got %c\n", type, dbus_message_iter_get_arg_type (&iter_variant));
+    g_warning (_("atspi_dbus_get_property: Wrong type: expected %s, got %c\n"), type, dbus_message_iter_get_arg_type (&iter_variant));
     goto done;
   }
   if (!strcmp (type, "(so)"))
@@ -984,12 +1160,13 @@ _atspi_dbus_get_property (gpointer obj, const char *interface, const char *name,
   else
   {
     dbus_message_iter_get_basic (&iter_variant, data);
-    dbus_message_unref (reply);
     if (type [0] == 's')
       *(char **)data = g_strdup (*(char **)data);
   }
   retval = TRUE;
 done:
+  if (reply)
+    dbus_message_unref (reply);
   return retval;
 }
 
@@ -998,25 +1175,42 @@ _atspi_dbus_send_with_reply_and_block (DBusMessage *message)
 {
   DBusMessage *reply;
   DBusError err;
+  AtspiApplication *app;
+  DBusConnection *bus;
 
+  app = get_application (dbus_message_get_destination (message));
+
+  if (app && !app->bus)
+    return NULL;	/* will fail anyway; app has been disposed */
+
+  bus = (app ? app->bus : _atspi_bus());
   dbus_error_init (&err);
-  /* TODO: Write this function; allow reentrancy */
-  reply = dbus_connection_send_with_reply_and_block (_atspi_bus(), message, 1000, &err);
+  reply = dbind_send_and_allow_reentry (bus, message, &err);
+  _atspi_process_deferred_messages ((gpointer)TRUE);
   dbus_message_unref (message);
   if (err.message)
-    g_warning ("Atspi: Got error: %s\n", err.message);
+  {
+    g_warning (_("AT-SPI: Got error: %s\n"), err.message);
+    dbus_error_free (&err);
+  }
   return reply;
 }
 
 GHashTable *
-_atspi_dbus_hash_from_message (DBusMessage *message)
+_atspi_dbus_return_hash_from_message (DBusMessage *message)
 {
   DBusMessageIter iter;
+  GHashTable *ret;
 
-  _ATSPI_DBUS_CHECK_SIG (message, "a{ss}", NULL);
+  if (!message)
+    return NULL;
+
+  _ATSPI_DBUS_CHECK_SIG (message, "a{ss}", NULL, NULL);
 
   dbus_message_iter_init (message, &iter);
-  return _atspi_dbus_hash_from_iter (&iter);
+  ret = _atspi_dbus_hash_from_iter (&iter);
+  dbus_message_unref (message);
+  return ret;
 }
 
 GHashTable *
@@ -1040,15 +1234,21 @@ _atspi_dbus_hash_from_iter (DBusMessageIter *iter)
 }
 
 GArray *
-_atspi_dbus_attribute_array_from_message (DBusMessage *message)
+_atspi_dbus_return_attribute_array_from_message (DBusMessage *message)
 {
   DBusMessageIter iter;
+  GArray *ret;
 
-  _ATSPI_DBUS_CHECK_SIG (message, "a{ss}", NULL);
+  if (!message)
+    return NULL;
+
+  _ATSPI_DBUS_CHECK_SIG (message, "a{ss}", NULL, NULL);
 
   dbus_message_iter_init (message, &iter);
 
-  return _atspi_dbus_attribute_array_from_iter (&iter);
+  ret = _atspi_dbus_attribute_array_from_iter (&iter);
+  dbus_message_unref (message);
+  return ret;
 }
 
 GArray *
@@ -1077,3 +1277,61 @@ _atspi_dbus_attribute_array_from_iter (DBusMessageIter *iter)
   return array;
 }
 
+void
+_atspi_dbus_set_interfaces (AtspiAccessible *accessible, DBusMessageIter *iter)
+{
+  DBusMessageIter iter_array;
+
+  accessible->interfaces = 0;
+  dbus_message_iter_recurse (iter, &iter_array);
+  while (dbus_message_iter_get_arg_type (&iter_array) != DBUS_TYPE_INVALID)
+  {
+    const char *iface;
+    gint n;
+    dbus_message_iter_get_basic (&iter_array, &iface);
+    if (!strcmp (iface, "org.freedesktop.DBus.Introspectable")) continue;
+    n = _atspi_get_iface_num (iface);
+    if (n == -1)
+    {
+      g_warning (_("AT-SPI: Unknown interface %s"), iface);
+    }
+    else
+      accessible->interfaces |= (1 << n);
+    dbus_message_iter_next (&iter_array);
+  }
+  accessible->cached_properties |= ATSPI_CACHE_INTERFACES;
+}
+
+void
+_atspi_dbus_set_state (AtspiAccessible *accessible, DBusMessageIter *iter)
+{
+  DBusMessageIter iter_array;
+  gint count;
+  dbus_uint32_t *states;
+
+  dbus_message_iter_recurse (iter, &iter_array);
+  dbus_message_iter_get_fixed_array (&iter_array, &states, &count);
+  if (count != 2)
+  {
+    g_warning (_("AT-SPI: expected 2 values in states array; got %d\n"), count);
+    if (!accessible->states)
+      accessible->states = _atspi_state_set_new_internal (accessible, 0);
+  }
+  else
+  {
+    guint64 val = ((guint64)states [1]) << 32;
+    val += states [0];
+    if (!accessible->states)
+      accessible->states = _atspi_state_set_new_internal (accessible, val);
+    else
+      accessible->states->states = val;
+  }
+  accessible->cached_properties |= ATSPI_CACHE_STATES;
+}
+
+GQuark
+atspi_error_quark (void)
+{
+  return g_quark_from_static_string ("atspi_error");
+}
+
